from pathlib import Path

import jinja2
import yaml  # type: ignore


def generate_conf(directory):
    HERE = Path(__file__).parent
    templateLoader = jinja2.FileSystemLoader(searchpath=HERE)
    templateEnv = jinja2.Environment(loader=templateLoader)
    TEMPLATE_FILE = "_template_conf.py"
    template = templateEnv.get_template(TEMPLATE_FILE)

    with open("./lamin-project.yaml") as f:
        try:
            variables = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            print(exc)
            quit()

    # see the cookiecutter.json!
    if "project_slug" not in variables:
        variables["project_slug"] = variables["project_name"].lower().replace(" ", "-")
    if "repository_name" not in variables:  # this differs from cookiecutter!
        variables["repository_name"] = variables["project_slug"]
    if "package_name" not in variables:
        variables["package_name"] = variables["project_slug"].lower().replace("-", "_")

    # prefix with `lamin_` and logg out
    variables_template = {}
    print("generating docs using:")
    for key, value in variables.items():
        print(f"    {key}: {value}")
        variables_template["lamin_" + key] = variables[key]
    print("\n")

    outputText = template.render(variables_template)

    with open(Path(directory) / "conf.py", "w") as f:
        f.write(outputText)

    return outputText



# -------------------------------------------------------------------------------------
# this renders the front matter

import yaml  # type: ignore
from myst_parser.mdit_to_docutils.base import (  # noqa
    DocutilsRenderer,
    SyntaxTreeNode,
    token_line,
)

from ._authors import authors


# from myst_parser 0.18.0
# https://github.com/executablebooks/MyST-Parser/blob/391a8cd1097db16f122ce4736e8924ecfb23e621/myst_parser/mdit_to_docutils/base.py#L792-L831  # noqa
def render_front_matter(self, token: SyntaxTreeNode) -> None:
    """Pass document front matter data."""
    position = token_line(token, default=0)

    if isinstance(token.content, str):
        try:
            data = yaml.safe_load(token.content)
        except (yaml.parser.ParserError, yaml.scanner.ScannerError):
            self.create_warning(
                "Malformed YAML",
                line=position,
                append_to=self.current_node,
                subtype="topmatter",
            )
            return
    else:
        data = token.content

    if not isinstance(data, dict):
        self.create_warning(
            f"YAML is not a dict: {type(data)}",
            line=position,
            append_to=self.current_node,
            subtype="topmatter",
        )
        return

    fields = {
        k: v
        for k, v in data.items()
        if k not in ("myst", "mystnb", "substitutions", "html_meta")
    }
    if fields:
        field_list = self.dict_to_fm_field_list(
            fields, language_code=self.document.settings.language_code
        )
        self.current_node.append(field_list)

    # end of copy of this function, the rest here is our code
    def format_date():
        if isinstance(data["date"], str):
            return data["date"].split(" ")[0]  # do not display time!
        else:
            return data["date"]

    html = ""
    # some posts might not have software, hence the · at the end
    if data.get("docs"):
        docs = f"<a href={data['docs']}>Documentation</a>"
        html += f"{docs}"
    if data.get("repo"):
        if html != "" and not html.endswith(" · "):
            html += " · "
        html += f"<a href={data['repo']}>Repository</a>"
    if data.get("tweet"):
        if html != "" and not html.endswith(" · "):
            html += " · "
        html += f"<a href={data['tweet']}>Tweet</a>"
    if data.get("linkedin"):
        if html != "" and not html.endswith(" · "):
            html += " · "
        html += f"<a href={data['linkedin']}>LinkedIn</a>"
    if data.get("doi"):
        if html != "" and not html.endswith(" · "):
            html += " · "
        html += f'<a href=https://doi.org/{data["doi"]}>DOI</a>'
    float_right = ""
    if data.get("number"):
        float_right += f'<li> ⸻ #{data["number"]}</li>'
    html = f"""<ul class="ablog-archive" style="padding-left: 0px"><li>{html}</li>{float_right}</ul>"""  # noqa
    self.nested_render_text(f"{html}", 0)

    if data.get("title") and self.md_config.title_to_header:
        self.nested_render_text(f"# {data['title']}", 0)

    def format_authors():
        if data.get("affiliation"):
            affiliation = data["affiliation"]
        else:
            affiliation = {}

        def format_title(k):
            return f'title="{affiliation[k]}"' if affiliation else ""

        return ", ".join(
            [
                f'<a href="{authors[k][1]}" {format_title(k)}>{authors[k][0]}</a>'
                for k in data["author"].split(", ")
            ]
        )

    if data.get("author"):
        html = f"{format_date()} · "
        html += f"{format_authors()}"
        html = f"""<ul class="ablog-archive" style="padding-left: 0px"><li>{html}</li></ul>"""  # noqa
        self.nested_render_text(f"{html}", 0)


DocutilsRenderer.render_front_matter = render_front_matter



nitpick_ignore = [
    ("py:class", "lnschema_core._core.dobject"),
    ("py:class", "lnschema_core._core.dtransform"),
    ("py:class", "lnschema_core._core.jupynb"),
    ("py:class", "pandas.core.frame.DataFrame"),
    ("py:class", "sqlmodel.orm.session.Session"),
    ("py:class", "numpy.ndarray"),
    ("py:class", "datetime.datetime"),
    ("py:class", "pydantic.main.BaseModel"),
    ("py:class", "cloudpathlib.cloudpath.CloudPath"),
    ("py:class", "anndata._core.anndata.AnnData"),
    ("py:class", "enum.Enum"),
    ("py:class", "pathlib.Path"),
    ("py:class", "Model"),
    ("py:class", "sqlmodel.main.SQLModel"),
    ("py:class", "MetaData"),
    ("py:class", "DictStrAny"),
    ("py:class", "unicode"),
    ("py:class", "typing.DictStrAny"),
    ("py:class", "typing.unicode"),
    ("py:data", "typing.Optional"),
    ("py:data", "typing.Literal"),
    ("py:data", "typing.Union"),
    ("py:data", "typing.Any"),
    ("py:data", "typing.Tuple"),
    ("py:obj", "bionty.Gene"),
    ("py:obj", "bionty.Species"),
]



import argparse
import sys
from pathlib import Path
from subprocess import call

import yaml  # type: ignore
from dirsync import sync

from lndocs._generate_conf import generate_conf

from ._postprocess import (
    replace_image_targets,
    replace_index_targets,
    update_index_file,
)

HERE = Path(__file__).parent


def get_lamin_project():
    with open("./lamin-project.yaml") as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as exc:
            print(exc)
            quit()


def main():
    parser = argparse.ArgumentParser(description="Build Lamin website.")
    aa = parser.add_argument
    aa("--show", action="store_true", help="launch server & show")
    aa("--docs", type=str, default="docs", help="directory with docs sources")
    aa("--site", type=str, default="_build/html", help="output directory")
    aa("--live", action="store_true", help="use autobuild")
    args = parser.parse_args()
    if not Path(args.docs).exists():
        sys.exit(
            f"The source directory {args.docs} does not exist! Change to repo root!"
        )
    if not Path("./lamin-project.yaml").exists():
        sys.exit("The ./lamin-project.yaml conf file does not exist!")
    lamin_project = get_lamin_project()

    sync(str(HERE / "lamin_sphinx"), "./lamin_sphinx", "sync", create=True, ctime=True)
    # check whether we need to generate the conf.py for Sphinx
    # input for it is the lamin-project.yaml file
    generate_conf_check = False
    conf_file = Path(args.docs) / "conf.py"
    if not conf_file.exists():
        generate_conf_check = True
    else:
        with open(conf_file) as f:
            first_line = f.readline()
        if first_line == "# auto-generated by lndocs\n":
            generate_conf_check = True
    if generate_conf_check:
        generate_conf(args.docs)

    if args.live:
        build_command = "sphinx-autobuild"
    else:
        build_command = "sphinx-build"

    # all of what follows here is about getting rid of the back-slash for index files
    # on URLs for the dedicated docs pages
    docs_dir = args.docs
    check_postprocess = not args.live and generate_conf_check
    notebooks = []
    if check_postprocess:
        docs_dir = Path(f"_{args.docs}_tmp/")
        sync(
            args.docs,
            Path(docs_dir) / lamin_project["project_slug"],
            "sync",
            create=True,
        )
        (docs_dir / lamin_project["project_slug"] / "conf.py").rename(
            docs_dir / "conf.py"
        )
        for path in docs_dir.glob("**/*"):
            if path.suffix not in {".md", ".ipynb"}:
                continue
            if ".ipynb_checkpoints/" in str(path):
                continue
            if path.suffix == ".ipynb":
                # test whether prefix is capital letter or digit and if so,
                # strip them for pretty & persistent urls we need the prefixes
                # on notebooks to allow users to navigate downloaded notebooks
                # that should display in order in a file browser
                prefix = path.stem[0]
                if prefix.isdigit() or prefix.isupper() and "-" in path.stem:
                    new_stem = "-".join(path.stem.split("-")[1:])
                    # path.with_stem() is >3.9
                    new_path = path.with_name(f"{new_stem}{path.suffix}")
                    path.rename(new_path)
                    path = new_path
                notebooks.append(path)
            if path.is_file():
                replace_index_targets(path, lamin_project["project_slug"])
            if str(path).endswith("index.md"):
                update_index_file(path)
        (docs_dir / f"{lamin_project['project_slug']}.md").rename(docs_dir / "index.md")
    # end of _docs_tmp

    build_status = call(f"{build_command} {docs_dir} {args.site}", shell=True)

    # delete auto-generated files
    if "package_name" in lamin_project:
        package_name = lamin_project["package_name"]
        for generated in Path(args.docs).glob(f"{package_name}.*.rst"):
            generated.unlink()

    site = Path(args.site)

    # move Jupyter-generated images to accessible location
    # as the root isn't part of the deployed website!
    if check_postprocess and (site / "_images").exists():
        for path in notebooks:
            path_html = (
                str(path).replace(str(docs_dir), str(site)).replace(".ipynb", ".html")
            )
            replace_image_targets(path_html, lamin_project["project_slug"])
        sync(
            site / "_images/",
            site / lamin_project["project_slug"] / "_images/",
            "sync",
            create=True,
        )

    if not args.show:
        if lamin_project["project_slug"] == "":  # deploy a copy of _static on Netlify
            sync(site / "_static", site / "docs/_static", "sync", create=True)
        return build_status
    else:
        if build_status == 0:
            call("cd _build/html; python -m http.server", shell=True)



# -------------------------------------------------------------------------------------
# this enables a footnote tooltip by setting the title element

from io import StringIO  # type: ignore  # noqa

from markdown import Markdown  # type: ignore  # noqa


def unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


# patching Markdown
Markdown.output_formats["plain"] = unmark_element
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def unmark(text):
    return __md.convert(text)


from docutils.nodes import footnote  # type: ignore  # noqa


def visit_footnote_reference(self, node):
    href = "#" + node["refid"]
    classes = "footnote-reference " + self.settings.footnote_references
    # walk through all nodes of the current document to find the
    # corresponding footnote and retrieve the text
    title = "See bottom of page."
    content = (
        node.document.children[0]
        if len(node.document.children[0]) > 1
        else node.document.children[1]
    )
    for node_ in content.children:
        # check whether a node is a footnote
        if isinstance(node_, footnote):
            print(node)
            print(node_.attributes["ids"])
            if node["refid"] in set(node_.attributes["ids"]):
                title = node_.children[1].rawsource
                break
        else:
            # repeat the same one level deeper in the tree
            if hasattr(node_, "children"):
                for node__ in node_.children:
                    if isinstance(node__, footnote):
                        if node["refid"] in set(node__.attributes["ids"]):
                            title = node__.children[1].rawsource
                            break
    if title == "See bottom of page.":
        print(f"WARNING: footnote text for footnote {node['refid']} not found")
    else:  # remove markup
        title = unmark(title)
    self.body.append(
        self.starttag(node, "a", "", CLASS=classes, href=href, title=title)
    )



authors = {
    "falexwolf": ("Alex Wolf", "https://falexwolf.me"),
    "sunnyosun": ("Sunny Sun", "https://github.com/sunnyosun"),
    "Koncopd": ("Sergei Rybakov", "https://github.com/Koncopd"),
    "Zethson": ("Lukas Heumos", "https://github.com/Zethson"),
}



import os
from datetime import datetime

import sphinx.ext.autosummary.generate
from docutils.writers._html_base import HTMLTranslator  # type: ignore  # noqa
from sphinx.application import Sphinx

author = "Lamin Labs"
copyright = f"{datetime.now():%Y}, {author}"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",  # needs to be after napoleon
    "IPython.sphinxext.ipython_console_highlighting",  # noqa https://github.com/spatialaudio/nbsphinx/issues/24
    "myst_nb",
    "ablog",
    "sphinxext.opengraph",
]

templates_path = ["../lamin_sphinx/_templates"]
source_suffix = [".rst", ".md", ".ipynb"]
exclude_patterns = [
    ".nox",
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "node_modules",
    "README.md",
    "**/README.md",
    "profile",
    "private",
]
default_role = "literal"
html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "show_prev_next": True,
    "use_edit_page_button": False,  # currently unused
    "search_bar_text": "Search",  # currently unused
    "navbar_end": ["theme-switcher"],
    "navbar_align": "left",
    "footer_items": ["copyright"],
    "pygment_light_style": "tango",  # https://help.farbox.com/pygments.html
    "pygment_dark_style": "monokai",
}

html_context = {
    "default_mode": "auto",
    "github_user": "laminlabs",
    "github_version": "main",
}

html_logo = (
    "https://raw.githubusercontent.com/laminlabs/lamin-profile/main/assets/logo.svg"
)
html_favicon = (
    "https://raw.githubusercontent.com/laminlabs/lamin-profile/main/assets/favicon.ico"
)
html_static_path = ["../lamin_sphinx/_static"]

# order matters below!
# https://stackoverflow.com/questions/45112812/sphinx-exclude-one-page-from-html-sidebars # noqa
html_sidebars = {
    "*": ["sidebar-nav-bs"],
    "**/*": ["sidebar-nav-bs"],
    "index": [],
    "impressum": [],
    "products": [],
    "docs": [],
}

# Other configurations
panels_add_bootstrap_css = False
myst_enable_extensions = [
    "deflist",
    "colon_fence",
]
myst_title_to_header = True  # allow frontmatter titles

autodoc_member_order = "bysource"
autodoc_typehints_format = "short"
napoleon_numpy_docstring = False
napoleon_use_rtype = True
napoleon_use_param = False

ogp_image = (
    "https://raw.githubusercontent.com/laminlabs/lamin-profile/main/assets/logo.svg"
)

# myst_nb options
nb_execution_mode = "off"

nitpicky = True  # report broken links

from . import _front_matter  # noqa
from ._authors import authors  # noqa
from ._cite_commands import register_cite  # noqa
from ._footnote_title import visit_footnote_reference  # noqa
from ._nitpick_ignore import nitpick_ignore  # noqa
from ._sort_autosummary import generate_autosummary_content  # noqa

HTMLTranslator.visit_footnote_reference = visit_footnote_reference
sphinx.ext.autosummary.generate.generate_autosummary_content = (
    generate_autosummary_content  # noqa
)


def setup(app: Sphinx):
    app.warningiserror = os.getenv("GITHUB_ACTIONS") is not None
    app.add_css_file("custom.css")
    app.connect("config-inited", register_cite)



from pathlib import Path


def replace_index_targets(filepath, project_slug):
    with open(filepath) as f:
        content = f.read()
    content = content.replace("/index", "")
    with open(filepath, "w") as f:
        f.write(content)


def replace_image_targets(filepath, project_slug):
    with open(filepath) as f:
        content = f.read()
    content = content.replace("_images/", f"{project_slug}/_images/")
    with open(filepath, "w") as f:
        f.write(content)


def update_index_file(filepath):
    with open(filepath) as f:
        content = f.read()

    Path(filepath).unlink()

    # fmt: off
    content = content\
        .replace("](", f"]({filepath.parent.stem}/")\
        .replace("}`", "}" + f"`{filepath.parent.stem}/")
    # fmt: on
    # revert for all http links
    content = content.replace(f"]({filepath.parent.stem}/http", "](http")

    # parse out indexed file list
    if "```{toctree}" in content:
        toctree = content.split("```{toctree}")[1]
        toctree = toctree.split("\n\n")[1]
        toctree = toctree.split("```")[0]

        toctree = toctree.strip("\n")

        toctree_split = toctree.split("\n")
        toctree_new = "\n".join(
            [f"{filepath.parent.stem}/{file}" for file in toctree_split]
        )

        content = content.replace(toctree, toctree_new)

    with open(f"{filepath.parent}.md", "w") as f:
        f.write(content)



# auto-generated by lndocs
import sys
from pathlib import Path

import {{ lamin_package_name }}

HERE = Path(__file__).parent
sys.path[:0] = [str(HERE), str(HERE.parent)]
from lamin_sphinx import *  # noqa
from lamin_sphinx import html_context, html_theme_options

project = "{{ lamin_project_name }}"
html_title = "{{ lamin_project_name }} | Lamin Labs"
release = {{ lamin_package_name }}.__version__
html_context["github_repo"] = "{{ lamin_repository_name }}"  # noqa
# We're actually using this for the link behind the brand of the page!
html_theme_options["logo"] = {"link": "docs/{{ lamin_project_slug }}"}  # noqa

ogp_site_url = "https://lamin.ai/docs/{{ lamin_project_slug }}"



# this modifies a function from sphinx.ext.autosummary.generate
# see the comment re where the modification happens
import pkgutil
from typing import Any, Dict, List, Set, Tuple

import sphinx
from sphinx.ext.autodoc.importer import import_module
from sphinx.ext.autosummary import get_documenter
from sphinx.ext.autosummary.generate import (
    AutosummaryRenderer,
    ModuleScanner,
    members_of,
    split_full_qualified_name,
)
from sphinx.pycode import ModuleAnalyzer, PycodeError
from sphinx.util import logging
from sphinx.util.inspect import safe_getattr

logger = logging.getLogger(__name__)


def generate_autosummary_content(
    name: str,
    obj: Any,
    parent: Any,
    template: AutosummaryRenderer,
    template_name: str,
    imported_members: bool,
    app: Any,
    recursive: bool,
    context: Dict,
    modname: str = None,
    qualname: str = None,
) -> str:
    doc = get_documenter(app, obj, parent)

    def skip_member(obj: Any, name: str, objtype: str) -> bool:
        try:
            return app.emit_firstresult(
                "autodoc-skip-member", objtype, name, obj, False, {}
            )
        except Exception as exc:
            logger.warning(
                "autosummary: failed to determine %r to be documented, "
                "the following exception was raised:\n%s",
                name,
                exc,
                type="autosummary",
            )
            return False

    # We only modified get_class_members by including the `order` variable!
    # This here was the original function
    # def get_class_members(obj: Any) -> Dict[str, Any]:
    #     members = sphinx.ext.autodoc.get_class_members(obj, [qualname], safe_getattr)
    #     return {name: member.object for name, member in members.items()}
    def get_class_members(obj: Any) -> Dict[str, Any]:
        members = sphinx.ext.autodoc.get_class_members(obj, [qualname], safe_getattr)  # type: ignore  # noqa
        if "__fields__" in members:
            order = members["__fields__"][1].keys()
        else:
            order = []
        all_members = {}
        for name in order:
            if name in members:
                all_members[name] = members[name].object
        for name, member in members.items():
            if name not in order:
                all_members[name] = member.object
        return all_members

    def get_module_members(obj: Any) -> Dict[str, Any]:
        members = {}
        for name in members_of(obj, app.config):
            try:
                members[name] = safe_getattr(obj, name)
            except AttributeError:
                continue
        return members

    def get_all_members(obj: Any) -> Dict[str, Any]:
        if doc.objtype == "module":
            return get_module_members(obj)
        elif doc.objtype == "class":
            return get_class_members(obj)
        return {}

    def get_members(
        obj: Any, types: Set[str], include_public: List[str] = [], imported: bool = True
    ) -> Tuple[List[str], List[str]]:
        items: List[str] = []
        public: List[str] = []

        all_members = get_all_members(obj)
        for name, value in all_members.items():
            documenter = get_documenter(app, value, obj)
            if documenter.objtype in types:
                # skip imported members if expected
                if imported or getattr(value, "__module__", None) == obj.__name__:
                    skipped = skip_member(value, name, documenter.objtype)
                    if skipped is True:
                        pass
                    elif skipped is False:
                        # show the member forcedly
                        items.append(name)
                        public.append(name)
                    else:
                        items.append(name)
                        if name in include_public or not name.startswith("_"):
                            # considers member as public
                            public.append(name)
        return public, items

    def get_module_attrs(members: Any) -> Tuple[List[str], List[str]]:
        """Find module attributes with docstrings."""
        attrs, public = [], []
        try:
            analyzer = ModuleAnalyzer.for_module(name)
            attr_docs = analyzer.find_attr_docs()
            for namespace, attr_name in attr_docs:
                if namespace == "" and attr_name in members:
                    attrs.append(attr_name)
                    if not attr_name.startswith("_"):
                        public.append(attr_name)
        except PycodeError:
            pass  # give up if ModuleAnalyzer fails to parse code
        return public, attrs

    def get_modules(obj: Any) -> Tuple[List[str], List[str]]:
        items: List[str] = []
        for _, modname, _ispkg in pkgutil.iter_modules(obj.__path__):
            fullname = name + "." + modname
            try:
                module = import_module(fullname)
                if module and hasattr(module, "__sphinx_mock__"):
                    continue
            except ImportError:
                pass

            items.append(fullname)
        public = [x for x in items if not x.split(".")[-1].startswith("_")]
        return public, items

    ns: Dict[str, Any] = {}
    ns.update(context)

    if doc.objtype == "module":
        scanner = ModuleScanner(app, obj)
        ns["members"] = scanner.scan(imported_members)
        ns["functions"], ns["all_functions"] = get_members(
            obj, {"function"}, imported=imported_members
        )
        ns["classes"], ns["all_classes"] = get_members(
            obj, {"class"}, imported=imported_members
        )
        ns["exceptions"], ns["all_exceptions"] = get_members(
            obj, {"exception"}, imported=imported_members
        )
        ns["attributes"], ns["all_attributes"] = get_module_attrs(ns["members"])
        ispackage = hasattr(obj, "__path__")
        if ispackage and recursive:
            ns["modules"], ns["all_modules"] = get_modules(obj)
    elif doc.objtype == "class":
        ns["members"] = dir(obj)
        ns["inherited_members"] = set(dir(obj)) - set(obj.__dict__.keys())
        ns["methods"], ns["all_methods"] = get_members(obj, {"method"}, ["__init__"])
        ns["attributes"], ns["all_attributes"] = get_members(
            obj, {"attribute", "property"}
        )
        print(ns["attributes"])

    if modname is None or qualname is None:
        modname, qualname = split_full_qualified_name(name)

    if doc.objtype in ("method", "attribute", "property"):
        ns["class"] = qualname.rsplit(".", 1)[0]  # type: ignore

    if doc.objtype in ("class",):
        shortname = qualname
    else:
        shortname = qualname.rsplit(".", 1)[-1]  # type: ignore

    ns["fullname"] = name
    ns["module"] = modname
    ns["objname"] = qualname
    ns["name"] = shortname

    ns["objtype"] = doc.objtype
    ns["underline"] = len(name) * "="

    if template_name:
        return template.render(template_name, ns)
    else:
        return template.render(doc.objtype, ns)



import nox

nox.options.reuse_existing_virtualenvs = True


@nox.session
def lint(session: nox.Session) -> None:
    session.install("pre-commit")
    session.run("pre-commit", "install")
    session.run("pre-commit", "run", "--all-files")


@nox.session
def build(session):
    session.install(".[dev,test]")
    session.run("pytest")



# -------------------------------------------------------------------------------------
# cite commands

from types import MappingProxyType  # noqa
from typing import Any, Mapping, NamedTuple, Sequence  # noqa

from docutils import nodes  # type:ignore # noqa
from docutils.parsers.rst.directives import class_option  # type:ignore # noqa
from docutils.parsers.rst.states import Inliner  # type:ignore # noqa
from sphinx.application import Sphinx  # noqa
from sphinx.config import Config  # noqa


class AutoLink(NamedTuple):
    class_name: str
    url_template: str
    title_template: str = "{}"  # noqa
    options: Mapping[str, Any] = MappingProxyType({"class": class_option})  # noqa

    def __call__(  # noqa
        self,
        name: str,
        rawtext: str,
        text: str,
        lineno: int,
        inliner: Inliner,
        options: Mapping[str, Any] = MappingProxyType({}),
        content: Sequence[str] = (),
    ):
        url = self.url_template.format(text)
        title = self.title_template.format(text)
        options = {**dict(classes=[self.class_name]), **options}
        node = nodes.reference(rawtext, title, refuri=url, **options)
        return [node], []


def register_cite(app: Sphinx, config: Config):
    # This follows the convention of natbib's \citet (Text) and \citep (Parantheses/Brackets)  # noqa
    app.add_role("ct", AutoLink("ct", "#{}", "{}"))
    app.add_role("cp", AutoLink("cp", "#{}", "[{}]"))
